@@ -15,7 +15,7 @@
 #   cd remote-desktop
 #   perl src/cgi-bin/qemu-web-desktop.pl --dir_service=src/html/desktop \
 #     --dir_html=src/html --dir_snapshots=/tmp \
-#     --dir_machines=src/html/desktop/machines/ 
+#     --dir_machines=src/html/desktop/machines/ --oneshot=1
 #
 # Then follow printed instructions in the terminal:
 # open a browser at something like:
@@ -72,7 +72,7 @@ use warnings qw( all );
 use CGI;                # use CGI.pm
 use File::Temp      qw/ tempdir tempfile /;
 use File::Path      qw/ rmtree  /;
-use File::Basename  qw(fileparse);
+use File::Basename  qw/ fileparse dirname /;
 use List::MoreUtils qw(uniq); # liblist-moreutils-perl
 use List::Util;
 use Sys::CPU;           # libsys-cpu-perl           for CPU::cpu_count
@@ -120,7 +120,15 @@ our %config;
 
 $config{version}                  = "23.02.15";  # year.month.day
 
-require '/etc/qemu-web-desktop/config.pl';
+my $dirname = dirname(__FILE__);
+
+if (-e '/etc/qemu-web-desktop/config.pl') {
+  require '/etc/qemu-web-desktop/config.pl';
+} elsif (-e "$dirname/../config.pl") {
+  require "$dirname/../config.pl";
+} else {
+  die "Can not find config.pl (e.g. /etc/qemu-web-desktop/config.pl or $dirname/../config.pl";
+}
 
 # search detached GPU (via vfio-pci). Only use video part, no audio.
 {
@@ -494,6 +502,7 @@ if ($need_fallback_server and length($session{fallback_servers})) {
   
   if ($fallback_server) {
     # update the fallback_servers in the POST message
+    print STDERR "$0: Warning: $error\n";
     print STDERR "$0: forwarding request to: $fallback_server\n";
     
     $q->param( fallback_servers => $session{fallback_servers} );
@@ -634,13 +643,20 @@ if (not $error) {
 } # Create snapshot
 
 # LAUNCH optional 'configuration' scripts (system wide, and user level)
-if (not $error and $config{config_script}) {
-  foreach my $c (@{ $config{config_script} }) {
-    session_run_script(\%config, \%session, $c);
+{
+  my $cmd = "";
+  if (not $error and $config{config_script}) {
+    foreach my $c (@{ $config{config_script} }) {
+      $cmd .= session_run_script(\%config, \%session, $c);
+    }
+  }
+  if (not $error and $session{config_script}) {
+    $cmd .= session_run_script(\%config, \%session, $session{config_script});
+  }
+  if ($cmd) {
+    # call virt customize with the list of all scripts to execute, in order
+    my $res = `virt-customize -a $session{snapshot} $cmd`;
   }
-}
-if (not $error and $session{config_script}) {
-  session_run_script(\%config, \%session, $session{config_script});
 }
 
 # LAUNCH CLONED VM -------------------------------------------------------------
@@ -801,10 +817,13 @@ if (not $error) {
       $session{pid_wait} = $proc_novnc->pid;
     }
   }
+  
   if ($config{service_port_multiple}) {
     $session{url} = "$config{http}://$session{server_name}:$session{port}/vnc.html?resize=scale&autoconnect=true&host=$session{server_name}&port=$session{port}";
   } else {
-    $session{url} = "$config{http}://$session{server_name}:$session{port}/vnc.html?resize=scale&autoconnect=true&path=?token=$session{token_vnc}";
+    # in case of an Apache2 port redirection, change ':$session{port}' into e.g. 
+    # the Apache2 Location (e.g. '/darts').
+    $session{url} = "$config{http}://$session{server_name}:$session{port}/vnc.html?resize=scale&autoconnect=true&path=?token=$session{token_vnc}&port=$session{port}";
   }
 }
 
@@ -1124,7 +1143,7 @@ sub session_test_script {
 # In addition, the script text (expression, file or URL) may contain the syntax:
 #   if(EXPR):
 # so that the script content is evaluated when EXPR is true.
-# The script content and EXPR may use the @USER@ @SESSION_NAME@ and @VM@ symbols,  
+# The script content and EXPR may use the @USER@ @PW@ @SESSION_NAME@ and @VM@ symbols,  
 # which are replaced by the user name, the session name and the VM name.
 sub session_run_script {
   my $config_ref  = shift;
@@ -1185,24 +1204,25 @@ sub session_run_script {
       if (not $TF) { return ""; }
       
       # store script in our 'snapshot' directory for local exec.
-      open my $fh, ">", "$session{dir_snapshot}/configuration_script";
-      print $fh "$content\n";
-      close $fh;
+      my ($script_handle, $script_name) = tempfile(TEMPLATE => "$session{dir_snapshot}/configuration_script_XXXXXXXX");
+      if (not $script_name) { return ""; }
+      print $script_handle "$content\n";
+      close $script_handle;
       
       # output script into apache2/log
-      print STDERR $0.": $session{name}: $session{date}: $session{user} uses config_script=$script -> $session{dir_snapshot}/configuration_script\n";
-      print STDERR "$content\n";
+      print STDERR $0.": $session{name}: $session{date}: $session{user} uses config_script=$script -> $script_name\n";
+#      print STDERR "$content\n";
       
       # call virt-customize: start VM, execute script, stop it. Only runs on stopped VMs.
       # virt-customize -a vm.qcow2 --run snapshot/script
-      my $cmd = "virt-customize -a $session{snapshot} ";
+      # my $cmd = "virt-customize -a $session{snapshot} ";
+      my $cmd= "";
       if ($virt_customize_commands == 0) {
-        $cmd .= "--firstboot $session{dir_snapshot}/configuration_script";
+        $cmd = " --firstboot $script_name";
       } else {
-        $cmd .= "--commands-from-file $session{dir_snapshot}/configuration_script";
+        $cmd = " --commands-from-file $script_name";
       }
-      my $res = `$cmd`;
-      return $res;
+      return $cmd;
     }
   }
 } # session_run_script
@@ -1224,6 +1244,7 @@ sub service_housekeeping {
   
   # clean:
   # - remove orphan snapshots (no corresponding JSON file)
+  # - remove orphan JSON      (no corresponding snapshots)
   # - remove snapshots that have gone above their lifetime
   my $now         = localtime();
   foreach my $snapshot (glob("$dir/$service"."_*")) {
@@ -1233,7 +1254,7 @@ sub service_housekeeping {
       
       if (not -e "$cfg/$snaphot_name.json") {
         # remove orphan $snapshot (no JSON)
-        print STDERR $0." [$now] housekeeping: $snapshot (left-over)\n";
+        print STDERR $0." [$now] housekeeping: $snapshot (left-over, no JSON)\n";
         rmtree( $snapshot ) || print STDERR "Failed removing $snapshot";
       } else {
         my $session_ref = session_load(\%config, "$cfg/$snaphot_name.json");
@@ -1249,6 +1270,20 @@ sub service_housekeeping {
         } else { rmtree( $snapshot ) || print STDERR "Failed removing $snapshot"; }
       }
     }
+  } # for snapshot
+  
+  # scan JSON files
+  foreach my $json (glob("$cfg/*.json")) {
+    
+    # session exists, load JSON
+    my $session_ref = session_load(\%config, $json);
+    if ($session_ref) {
+      my %session = %{ $session_ref };
+      if (not -d $session{dir_snapshot}) { 
+        print STDERR $0." [$now] housekeeping: $json (left-over, no snapshot)\n";
+        unlink($json); 
+      }          # snapshot does not exist
+    }
   }
   
   # now count how many active sessions we have.
@@ -1704,9 +1739,12 @@ sub session_check_ldap {
 # specify an authenticator function in /etc/qemu-web-desktop/config.pl such as:
 #
 #  sub check_user_func {
-#    my $user = shift;
-#    my $pw   = shift;
-#    my $auth = shift; # previous auth
+#    my $user          = shift;
+#    my $pw            = shift;
+#    my $authenticated = shift; # previous authenticator results
+#    my $session_ref   = shift;  
+#    if (not $session_ref) { return $authenticated; }
+#    my %session       = %{ $session_ref };
 #    return "$auth and SUCCESS: [Custom] $user authenticated.";
 #  }
 #
